var path = require('path');
var pixelmatch = require('pixelmatch');
var fs = require('fs');
var pngjs = require('pngjs');
var moveFile = require('move-file');
var sharp = require('sharp');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

var path__default = /*#__PURE__*/_interopDefaultLegacy(path);
var pixelmatch__default = /*#__PURE__*/_interopDefaultLegacy(pixelmatch);
var fs__default = /*#__PURE__*/_interopDefaultLegacy(fs);
var moveFile__default = /*#__PURE__*/_interopDefaultLegacy(moveFile);
var sharp__default = /*#__PURE__*/_interopDefaultLegacy(sharp);

const PLUGIN_NAME = "cp-visual-regression-diff";
const IMAGE_SNAPSHOT_PREFIX = `__${PLUGIN_NAME}_snapshots__`;
var FILE_SUFFIX;

(function (FILE_SUFFIX) {
  FILE_SUFFIX["diff"] = ".diff";
  FILE_SUFFIX["actual"] = ".actual";
})(FILE_SUFFIX || (FILE_SUFFIX = {}));

const TASK = {
  getScreenshotPath: `${PLUGIN_NAME}-getScreenshotPath`,
  compareImages: `${PLUGIN_NAME}-compareImages`,
  approveImage: `${PLUGIN_NAME}-approveImage`,
  doesFileExist: `${PLUGIN_NAME}-doesFileExist`
};

function isHighSurrogate(codePoint) {
  return codePoint >= 0xd800 && codePoint <= 0xdbff;
}

function isLowSurrogate(codePoint) {
  return codePoint >= 0xdc00 && codePoint <= 0xdfff;
}

// Truncate string by size in bytes
var truncate = function truncate(getLength, string, byteLength) {
  if (typeof string !== "string") {
    throw new Error("Input must be string");
  }

  var charLength = string.length;
  var curByteLength = 0;
  var codePoint;
  var segment;

  for (var i = 0; i < charLength; i += 1) {
    codePoint = string.charCodeAt(i);
    segment = string[i];

    if (isHighSurrogate(codePoint) && isLowSurrogate(string.charCodeAt(i + 1))) {
      i += 1;
      segment += string[i];
    }

    curByteLength += getLength(segment);

    if (curByteLength === byteLength) {
      return string.slice(0, i + 1);
    }
    else if (curByteLength > byteLength) {
      return string.slice(0, i - segment.length + 1);
    }
  }

  return string;
};

var getLength = Buffer.byteLength.bind(Buffer);
var truncateUtf8Bytes = truncate.bind(null, getLength);

/*jshint node:true*/

/**
 * Replaces characters in strings that are illegal/unsafe for filenames.
 * Unsafe characters are either removed or replaced by a substitute set
 * in the optional `options` object.
 *
 * Illegal Characters on Various Operating Systems
 * / ? < > \ : * | "
 * https://kb.acronis.com/content/39790
 *
 * Unicode Control codes
 * C0 0x00-0x1f & C1 (0x80-0x9f)
 * http://en.wikipedia.org/wiki/C0_and_C1_control_codes
 *
 * Reserved filenames on Unix-based systems (".", "..")
 * Reserved filenames in Windows ("CON", "PRN", "AUX", "NUL", "COM1",
 * "COM2", "COM3", "COM4", "COM5", "COM6", "COM7", "COM8", "COM9",
 * "LPT1", "LPT2", "LPT3", "LPT4", "LPT5", "LPT6", "LPT7", "LPT8", and
 * "LPT9") case-insesitively and with or without filename extensions.
 *
 * Capped at 255 characters in length.
 * http://unix.stackexchange.com/questions/32795/what-is-the-maximum-allowed-filename-and-folder-size-with-ecryptfs
 *
 * @param  {String} input   Original filename
 * @param  {Object} options {replacement: String | Function }
 * @return {String}         Sanitized filename
 */



var illegalRe = /[\/\?<>\\:\*\|"]/g;
var controlRe = /[\x00-\x1f\x80-\x9f]/g;
var reservedRe = /^\.+$/;
var windowsReservedRe = /^(con|prn|aux|nul|com[0-9]|lpt[0-9])(\..*)?$/i;
var windowsTrailingRe = /[\. ]+$/;

function sanitize(input, replacement) {
  if (typeof input !== 'string') {
    throw new Error('Input must be string');
  }
  var sanitized = input
    .replace(illegalRe, replacement)
    .replace(controlRe, replacement)
    .replace(reservedRe, replacement)
    .replace(windowsReservedRe, replacement)
    .replace(windowsTrailingRe, replacement);
  return truncateUtf8Bytes(sanitized, 255);
}

var sanitizeFilename = function (input, options) {
  var replacement = (options && options.replacement) || '';
  var output = sanitize(input, replacement);
  if (replacement === '') {
    return output;
  }
  return sanitize(output, '');
};

const createImageResizer = (width, height) => source => {
  const resized = new pngjs.PNG({
    width,
    height,
    fill: true
  });
  pngjs.PNG.bitblt(source, resized, 0, 0, source.width, source.height, 0, 0);
  return resized;
};

const fillSizeDifference = (width, height) => image => {
  const inArea = (x, y) => y > height || x > width;

  for (let y = 0; y < image.height; y++) {
    for (let x = 0; x < image.width; x++) {
      if (inArea(x, y)) {
        const idx = image.width * y + x << 2;
        image.data[idx] = 0;
        image.data[idx + 1] = 0;
        image.data[idx + 2] = 0;
        image.data[idx + 3] = 64;
      }
    }
  }

  return image;
};

const importAndScaleImage = async cfg => {
  const imgBuffer = fs__default["default"].readFileSync(cfg.path);
  const rawImgNew = pngjs.PNG.sync.read(imgBuffer);
  if (cfg.scaleFactor === 1) return rawImgNew;
  const newImageWidth = Math.ceil(rawImgNew.width * cfg.scaleFactor);
  const newImageHeight = Math.ceil(rawImgNew.height * cfg.scaleFactor);
  await sharp__default["default"](imgBuffer).resize(newImageWidth, newImageHeight).toFile(cfg.path);
  return pngjs.PNG.sync.read(fs__default["default"].readFileSync(cfg.path));
};

const alignImagesToSameSize = (firstImage, secondImage) => {
  const firstImageWidth = firstImage.width;
  const firstImageHeight = firstImage.height;
  const secondImageWidth = secondImage.width;
  const secondImageHeight = secondImage.height;
  const resizeToSameSize = createImageResizer(Math.max(firstImageWidth, secondImageWidth), Math.max(firstImageHeight, secondImageHeight));
  const resizedFirst = resizeToSameSize(firstImage);
  const resizedSecond = resizeToSameSize(secondImage);
  return [fillSizeDifference(firstImageWidth, firstImageHeight)(resizedFirst), fillSizeDifference(secondImageWidth, secondImageHeight)(resizedSecond)];
};

const getConfigVariableOrThrow = (config, name) => {
  if (config[name]) return config[name];
  throw `[Image snapshot] CypressConfig.${name} cannot be missing or \`false\`!`;
};

const initPlugin = (on, config) => {
  if (config.env["pluginVisualRegressionForceDeviceScaleFactor"] !== false) {
    // based on https://github.com/cypress-io/cypress/issues/2102#issuecomment-521299946
    on("before:browser:launch", (browser, launchOptions) => {
      if (browser.name === "chrome" || browser.name === "chromium") {
        launchOptions.args.push("--force-device-scale-factor=1");
        launchOptions.args.push("--high-dpi-support=1");
      } else if (browser.name === "electron" && browser.isHeaded) {
        // eslint-disable-next-line no-console
        console.log("There isn't currently a way of setting the device scale factor in Cypress when running headed electron so we disable the image regression commands.");
      }
    });
  }

  on("task", {
    [TASK.getScreenshotPath]({
      title,
      imagesDir,
      specPath
    }) {
      return path__default["default"].join(IMAGE_SNAPSHOT_PREFIX, path__default["default"].dirname(specPath), ...imagesDir.split("/"), `${sanitizeFilename(title)}${FILE_SUFFIX.actual}.png`);
    },

    [TASK.doesFileExist]({
      path
    }) {
      return fs__default["default"].existsSync(path);
    },

    [TASK.approveImage]({
      img
    }) {
      const oldImg = img.replace(FILE_SUFFIX.actual, "");
      if (fs__default["default"].existsSync(oldImg)) fs__default["default"].unlinkSync(oldImg);
      const diffImg = img.replace(FILE_SUFFIX.actual, FILE_SUFFIX.diff);
      if (fs__default["default"].existsSync(diffImg)) fs__default["default"].unlinkSync(diffImg);
      if (fs__default["default"].existsSync(img)) moveFile__default["default"].sync(img, oldImg);
      return null;
    },

    async [TASK.compareImages](cfg) {
      let imgDiff;
      let errorMsg;

      if (fs__default["default"].existsSync(cfg.imgOld) && !cfg.updateImages) {
        const rawImgNew = await importAndScaleImage({
          scaleFactor: cfg.scaleFactor,
          path: cfg.imgNew
        });
        const rawImgOld = pngjs.PNG.sync.read(fs__default["default"].readFileSync(cfg.imgOld));
        const isImgSizeDifferent = rawImgNew.height !== rawImgOld.height || rawImgNew.width !== rawImgOld.width;
        const [imgNew, imgOld] = isImgSizeDifferent ? alignImagesToSameSize(rawImgNew, rawImgOld) : [rawImgNew, rawImgOld];
        const {
          width,
          height
        } = imgNew;
        const diff = new pngjs.PNG({
          width,
          height
        });
        const diffConfig = Object.assign({
          includeAA: true
        }, cfg.diffConfig);
        const diffPixels = pixelmatch__default["default"](imgNew.data, imgOld.data, diff.data, width, height, diffConfig);
        imgDiff = diffPixels / (width * height);

        if (isImgSizeDifferent) {
          errorMsg = `Images size mismatch - new screenshot is ${rawImgNew.width}px by ${rawImgNew.height}px while old one is ${rawImgOld.width}px by ${rawImgOld.height} (width x height).`;
        } else if (imgDiff > cfg.maxDiffThreshold) {
          const roundedImgDiff = Math.ceil(imgDiff * 1000) / 1000;
          errorMsg = `Image diff factor (${roundedImgDiff}) is bigger than maximum threshold option ${cfg.maxDiffThreshold}`;
        }

        if (errorMsg) {
          fs__default["default"].writeFileSync(cfg.imgNew.replace(FILE_SUFFIX.actual, FILE_SUFFIX.diff), pngjs.PNG.sync.write(diff));
          return {
            error: true,
            message: errorMsg,
            imgDiff,
            maxDiffThreshold: cfg.maxDiffThreshold
          };
        } else {
          // don't overwrite file if it's the same (imgDiff < cfg.maxDiffThreshold && !isImgSizeDifferent)
          fs__default["default"].unlinkSync(cfg.imgNew);
        }
      } else {
        // there is no "old screenshot" or screenshots should be immediately updated
        imgDiff = 0;
        moveFile__default["default"].sync(cfg.imgNew, cfg.imgOld);
      }

      if (typeof imgDiff !== "undefined") {
        const roundedImgDiff = Math.ceil(imgDiff * 1000) / 1000;
        return {
          message: `Image diff (${roundedImgDiff}%) is within boundaries of maximum threshold option ${cfg.maxDiffThreshold}`,
          imgDiff,
          maxDiffThreshold: cfg.maxDiffThreshold
        };
      }

      return null;
    }

  });
  on("after:screenshot", details => {
    var _details$name;

    if (((_details$name = details.name) == null ? void 0 : _details$name.indexOf(IMAGE_SNAPSHOT_PREFIX)) !== 0) return;
    return new Promise((resolve, reject) => {
      const screenshotsFolder = getConfigVariableOrThrow(config, "screenshotsFolder");
      const newRelativePath = details.name.substring(IMAGE_SNAPSHOT_PREFIX.length + path__default["default"].sep.length);
      const newAbsolutePath = path__default["default"].normalize(path__default["default"].join(config.projectRoot, newRelativePath));
      void moveFile__default["default"](details.path, newAbsolutePath).then(() => {
        fs__default["default"].rm(path__default["default"].join(screenshotsFolder, IMAGE_SNAPSHOT_PREFIX), {
          recursive: true,
          force: true
        }, err => {
          if (err) return reject(err);
          resolve({
            path: newAbsolutePath
          });
        });
      }).catch(reject);
    });
  });
};

exports.initPlugin = initPlugin;
//# sourceMappingURL=plugins.js.map
